{******************************************************************************}
{* This code was written by : Jamie Hart                                      *}
{******************************************************************************}
{* TXBase is a Non-visual component for Delphi 2/3 which gives access to      *}
{* dBase III+ data files without needing the BDE installed.                   *}
{*                                                                            *}
{* This component is freeware: any comments or suggestions are welcome at:    *}
{* jay@bitsmart.com                                                           *}
{*                                                                            *}
{* Version 0.99a (beta)                                                       *}
{******************************************************************************}
{* Users of the TXBase component must accept the following disclaimer of      *}
{* warranty:                                                                  *}
{*                                                                            *}
{* TXBase is supplied as is. The author disclaims all warranties, expressed   *}
{* or implied, including, without limitation, the warranties of               *}
{* merchantability and of fitness for any purpose. The author assumes no      *}
{* liability for damages, direct or consequential, which may result from the  *}
{* use of TXBase.                                                             *}
{******************************************************************************}

Unit XBase;

Interface

Uses
  Windows, SysUtils, Classes, DsgnIntf, Dialogs;

Type
  EXBaseException = Class(Exception);

Type
  TXBFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;


Type
  THeader = Record
    VersionNumber: Byte;
    LastUpdated: Array[1..3] Of Byte;
    NumRecords: DWORD;
    HeaderLen: Word;
    RecordLen: Word;
    Reserved: Word;
    IncTrans: Byte;
    EncFlag: Byte;
    Reserved2: Array[1..12] Of Byte;
    MdxFlag: Byte;
    LangDriver: Byte;
    Reserved3: Word;
  End;
  
Type
  TFieldStruct = Record
    FieldName: Array[1..11] Of Char;
    FieldType: Char;
    Address: DWORD;
    FieldLength: Byte;
    Decimals: Byte;
    Reserved: Word;
    WorkArea: Byte;
    Reserved2: Word;
    SetFields: Byte;
    Reserved3: Array[1..7] Of Byte;
    IndexFlag: Byte;
  End;
  
Type
  TDBTHeader = Record
    NextBlock: DWORD;
    BlockSize: DWORD;
    Reserved: Array[1..504] Of Char;
  End;
  
Type
  TXBChangeEvent = Procedure (Sender: TObject; Var Allowed: Boolean) Of Object;
  TXBDeleteEvent = Procedure (Sender: TObject; Var Allowed: Boolean) Of Object;
  
Type
  TFieldType = (xbfUnknown, xbfChar, xbfDate, xbfNumber, xbfLogic, xbfMemo);
  
Type
  TXBase = Class(TComponent)
  Private
    { Private Declarations }
    FFileName: String;
    FActive: Boolean;
    FModified: Boolean;
    FCurrentRecord: Longint;
    FOnChanging: TXBChangeEvent;
    FOnDeleting: TXBDeleteEvent;
    FOnAppended: TNotifyEvent;
    NumFields: Integer;
    IsOpen: Boolean;
    Header: THeader;
    DBFile: TFileStream;
    MemoFile: TFileStream;
    FieldStruct: Array[1..128] Of TFieldStruct;
    RecordBuffer: Array[1..4000] Of Char;
    DBTHeader: TDBTHeader;
    Function GetFileName: String;
    Procedure SetFileName(Name: String);
    Procedure SetActive(Val: Boolean);
    Procedure FillHeaderInfo;
    Procedure ReadARecord;
    Procedure WriteARecord;
    Function GetMemoData(FieldNo: Longint): String;
    Function GetFieldNumberFromName(Text: String): Integer;
    Function TrimSpaces(Text: String): String;
    Function PadLeft(Text: String; Len: Integer): String;
    Function PadRight(Text: String; Len: Integer): String;
    Procedure UpdateFieldData(FieldNo: integer; Text: String);
    Procedure UpdateMemoData(FieldNo: integer; Text: String);
    Function Changing: Boolean;
    Function DoDelete: Boolean;
    Function IsBof: Boolean;
    Function IsEof: Boolean;
    Function GetDeleteStatus: boolean;
    Procedure SetDeleteStatus(Delete: boolean);
  Protected
    { Protected Declarations }
  Public
    { Public Declarations }
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Function GetFieldByName(FieldName: String): String;
    Function GetFieldByNumber(FieldNum: Integer): String;
    Procedure GotoBOF;
    Procedure GotoEOF;
    Procedure GotoNext;
    Procedure GotoPrev;
    Procedure GotoRecord(Recnum: Longint);
    Procedure RefreshRecord;
    Function GetFieldName(FieldNo: Integer): String;
    Function GetFieldType(FieldNo: Integer): TFieldType;
    Function GetFieldSize(FieldNo: Integer): Integer;
    Procedure PutFieldByName(FieldName, Data: String);
    Procedure PutFieldByNumber(FieldNum: Integer; Data: String);
    Procedure PostChanges;
    Procedure AppendBlank;
  Published
    Property Filename: String Read GetFileName Write SetFileName;
    Property Active: Boolean Read FActive Write SetActive;
    Property Recno: Longint Read FCurrentRecord;
    Property RecordCount: Integer Read Header.NumRecords;
    Property FieldCount: Integer Read NumFields;
    Property OnChanging: TXBChangeEvent Read FOnChanging Write FOnChanging;
    Property OnDeleting: TXBDeleteEvent Read FOnDeleting Write FOnDeleting;
    Property OnAppended: TNotifyEvent read FOnAppended write FOnAppended;
    Property Bof: Boolean Read IsBof;
    Property Eof: Boolean Read IsEof;
    Property Deleted: Boolean Read GetdeleteStatus Write SetDeleteStatus;
    Property Modified: boolean Read FModified;
  End;

Procedure Register;

Implementation

Const
  dB3     = $03;
  dB4     = $04;
  dB5     = $05;
  dB3Memo = $83;
  dB4Memo = $84;

Constructor TXBase.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  IsOpen := False;
End;

Destructor TXBase.Destroy;
Begin
  Inherited Destroy;
End;

Function TXBase.TrimSpaces(Text: String): String;
Var
  len: Integer;
  S  : String; 
  
Begin
  S := Text;
  Len := Length(Text);
  While (Len > 0) And (Text[Len] = #32) Do Begin
    Delete(Text, Len, 1);
    Len := Len - 1;
  End;
  Result := Text;
End;

Procedure TXBase.FillHeaderInfo;
Var
  TmpC,
    TmpA: Integer;
  
Begin
  {Get Header Info}
  Try
    DBFile.Seek(0, soFromBeginning);
    DBFile.ReadBuffer(Header, Sizeof(Header));
    NumFields := Trunc((Header.HeaderLen - 34) / 32);
    
    {GetFieldInfo}
    TmpA := 2;
    For TmpC := 1 To NumFields Do Begin
      DBFile.Seek(TmpC * 32, soFromBeginning);
      DBFile.ReadBuffer(FieldStruct[TmpC], Sizeof(FieldStruct[TmpC]));
      FieldStruct[TmpC].Address := TmpA;
      TmpA := TmpA + FieldStruct[TmpC].FieldLength;
    End;
    FCurrentRecord := 1;
    ReadARecord;
  Except
    Raise EXBaseException.Create('Error Reading Header Info.');
  End;
End;

Procedure TXBase.ReadARecord;
Var
  TmpS: String;

Begin
  Try
    DBFile.Seek((Header.HeaderLen) + ((FCurrentRecord - 1) * Header.RecordLen), soFromBeginning);
    DBFile.ReadBuffer(RecordBuffer, Header.RecordLen);
    FModified := False;
  Except
    Raise EXBaseException.Create('Error Reading Record');
  End;
End;

Procedure TXBase.WriteARecord;
Var
  TmpS: String;
  
Begin
  Try
    DBFile.Seek((Header.HeaderLen) + ((FCurrentRecord - 1) * Header.RecordLen), soFromBeginning);
    DBFile.WriteBuffer(RecordBuffer, Header.RecordLen);
    FModified := False;
  Except
    Raise EXBaseException.Create('Error Writing Record');
  End;
End;

Procedure TXBase.PostChanges;
Begin
  WriteARecord;
End;

Function TXBase.GetFieldNumberFromName(Text: String): Integer;
Var
  TmpI: integer;
  
Begin
  TmpI := 1;
  While (GetFieldName(TmpI) <> UpperCase(Text)) And (TmpI <= NumFields) Do Begin
    inc(TmpI);
  End;
  If GetFieldName(TmpI) = UpperCase(Text) Then Begin
    Result := TmpI;
  End Else Begin
    Result := 0;
  End;
End;

Function TXBase.PadLeft(Text: String; Len: Integer): String;
Begin
  While Length(Text) < Len Do Begin
    Text := ' ' + Text;
  End;
  Result := Text;
End;

Function TXBase.PadRight(Text: String; Len: Integer): String;
Begin
  While Length(Text) < Len Do Begin
    Text := Text + ' ';
  End;
  Result := Text;
End;

Procedure TXBase.UpdateFieldData(FieldNo: integer; Text: String);
Var
  TmpL: Longint;
  
Begin
  If (FieldNo >= 1) And (FieldNo <= NumFields) Then Begin
    Text := TrimRight(Text);
    If (Length(Text) > 0) And (Length(Text) <= FieldStruct[FieldNo].FieldLength) Then Begin
      With FieldStruct[FieldNo] Do Begin
        If (FieldType = 'N') Or (FieldType = 'M') Then Begin
          Text := PadLeft(Text, FieldLength);
        End Else Begin
          Text := PadRight(Text, FieldLength);
        End;
        For TmpL := 1 To Length(Text) Do Begin
          RecordBuffer[Address - 1 + TmpL] := Text[TmpL];
        End;
      End;
      FModified := True;
    End Else Begin
      Raise EXBaseException.Create('Data Overflow');
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number');
  End;
End;

Procedure TXBase.UpdateMemoData(FieldNo: integer; Text: String);
Var
  TmpL : Longint;
  Block: String; 
  
Begin
  If (FieldNo >= 1) And (FieldNo <= NumFields) Then Begin
    Block := IntToStr(DBTHeader.NextBlock);
    Block := TrimLeft(TrimRight(Block));
    UpdateFieldData(FieldNo, Block);
    MemoFile.Seek( - 1, soFromEnd);
    Block := Chr(26) + Chr(26);
    For TmpL := 1 To Length(Block) Do Begin
      MemoFile.Write (Block[TmpL], 1);
    End;
    MemoFile.Seek(DBTHeader.NextBlock * 512, soFromBeginning);
    Text := Text + chr(26) + Chr(26);
    For TmpL := 1 To Length(Text) Do Begin
      MemoFile.Write (Text[TmpL], 1);
    End;
    inc(DBTHeader.NextBlock);
    FModified := True;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number');
  End;
End;

Function TXBase.GetMemoData(FieldNo: Longint): String;
Var
  TmpS: String;
  Buff: Array[1..512] Of Char;
  Flag: Boolean;

Begin
  Try
    TmpS := '';
    Flag := False;
    MemoFile.Seek(FieldNo * 512, soFromBeginning);
    While (Not Flag) Do Begin
      If MemoFile.Read (Buff, 512) < 512 Then Begin
        Flag := True;
      End;
      If Pos(#26#26, Buff) > 0 Then Begin
        TmpS := TmpS + Copy(Buff, 1, Pos(#26#26, Buff) - 1);
        Flag := True;
      End Else Begin
        TmpS := TmpS + Copy(Buff, 1, Pos(#26, Buff) - 1);
      End;
    End;
    While Pos(#141, TmpS) > 0 Do Begin
      TmpS[Pos(#141, TmpS)] := #13;
    End;
    Result := TmpS;
  Except
    Raise EXBaseException.Create('Error Reading Memo Record');
  End;
End;

Function TXBase.GetFieldByName(FieldName: String): String;
Var
  FieldNum: Integer;
  TmpS    : String; 
  
Begin
  FieldNum := GetFieldNumberFromName(FieldName);
  If FieldNum > 0 Then Begin
    If GetFieldType(FieldNum) <> xbfMemo Then Begin
      TmpS := Copy(RecordBuffer, FieldStruct[FieldNum].Address, FieldStruct[FieldNum].FieldLength);
      TmpS := TrimSpaces(TmpS);
    End Else Begin
      TmpS := Copy(RecordBuffer, FieldStruct[FieldNum].Address, FieldStruct[FieldNum].FieldLength);
      If StrToIntDef(TmpS, 0) > 0 Then Begin
        TmpS := GetMemoData(StrToIntDef(TmpS, 0));
      End Else Begin
        TmpS := '';
      End;
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Name');
  End;
  Result := TmpS;
End;

Function TXBase.GetFieldByNumber(FieldNum: Integer): String;
Var
  TmpS: String;

Begin
  If (FieldNum >= 1) And (FieldNum <= NumFields) Then Begin
    If GetFieldType(FieldNum) <> xbfMemo Then Begin
      TmpS := Copy(RecordBuffer, FieldStruct[FieldNum].Address, FieldStruct[FieldNum].FieldLength);
      TmpS := TrimSpaces(TmpS);
    End Else Begin
      TmpS := Copy(RecordBuffer, FieldStruct[FieldNum].Address, FieldStruct[FieldNum].FieldLength);
      If StrToIntDef(TmpS, 0) > 0 Then Begin
        TmpS := GetMemoData(StrToIntDef(TmpS, 0));
      End Else Begin
        TmpS := '';
      End;
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number');
  End;
  Result := TmpS;
End;

Procedure TXBase.PutFieldByName(FieldName, Data: String);
Var
  FieldNum: Integer;
  
Begin
  FieldNum := GetFieldNumberFromName(FieldName);
  If FieldNum > 0 Then Begin
    If FieldStruct[FieldNum].FieldType = 'M' Then Begin
      UpdateMemoData(FieldNum, Data);
    End Else Begin
      UpdateFieldData(FieldNum, Data);
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Name');
  End;
End;

Procedure TXBase.PutFieldByNumber(FieldNum: Integer; Data: String);
Begin
  If (FieldNum > 0) And (FieldNum <= NumFields) Then Begin
    If FieldStruct[FieldNum].FieldType = 'M' Then Begin
      UpdateMemoData(FieldNum, Data);
    End Else Begin
      UpdateFieldData(FieldNum, Data);
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number');
  End;
End;

Procedure TXBase.GotoBOF;
Begin
  If IsOpen Then Begin
    If Changing Then Begin
      FCurrentRecord := 1;
      ReadARecord;
    End;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Procedure TXBase.GotoEOF;
Begin
  If IsOpen Then Begin
    If Changing Then Begin
      FCurrentRecord := Header.NumRecords;
      ReadARecord;
    End;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Procedure TXBase.GotoNext;
Begin
  If IsOpen Then Begin
    If FCurrentRecord < Ord(Header.NumRecords) Then Begin
      If Changing Then Begin
        FCurrentRecord := FCurrentRecord + 1;
        ReadARecord;
      End;
    End;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Procedure TXBase.GotoPrev;
Begin
  If IsOpen Then Begin
    If FCurrentRecord > 1 Then Begin
      If Changing Then Begin
        FCurrentRecord := FCurrentRecord - 1;
        ReadARecord;
      End;
    End;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Procedure TXBase.GotoRecord(Recnum: Longint);
Begin
  If IsOpen Then Begin
    If (Recnum > 0) And (Recnum <= Header.NumRecords) Then Begin
      If Changing Then Begin
        FCurrentRecord := RecNum;
        ReadARecord;
      End;
    End Else Begin
      Raise EXBaseException.Create('Record Number Out of Range');
    End;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Procedure TXBase.RefreshRecord;
Begin
  If IsOpen Then Begin
    if Changing then Begin
      ReadARecord;
    end;
  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Function TXBase.GetFileName: String;
Begin
  Result := FFileName + '.dbf';
End;

Procedure TXBase.SetFileName(Name: String);
Begin
  FFileName := Name;
  If Pos('.', FFileName) > 0 Then Begin
    FfileName := Copy(FFilename, 1, Pos('.', FFileName) - 1);
  End;
  FActive := (FActive And FileExists(FFileName + '.dbf'));
  If FActive And (Not IsOpen) Then Begin
    DBFile := TFileStream.Create(FFileName + '.dbf', fmOpenReadWrite);
    IsOpen := True;
    FillHeaderInfo;
    If (Header.versionNumber = dB3Memo) Or (Header.versionNumber = dB4Memo) Then Begin
      MemoFile := TFileStream.Create(FFileName + '.dbt', fmOpenReadWrite);
      MemoFile.Seek(0, soFromBeginning);
      MemoFile.ReadBuffer(DBTHeader, Sizeof(DBTHeader));
    End;
  End;
  If (Not FActive) And (IsOpen) Then Begin
    DBFile.Free;
    MemoFile.Free;
    IsOpen := False;
  End;
End;

Procedure TXBase.SetActive(Val: Boolean);
Begin
  FActive := Val;
  FActive := (FActive And FileExists(FFileName + '.dbf'));
  If FActive And (Not IsOpen) Then Begin
    DBFile := TFileStream.Create(FFileName + '.dbf', fmOpenReadWrite);
    IsOpen := True;
    FillHeaderInfo;
    If (Header.versionNumber = dB3Memo) Or (Header.versionNumber = dB4Memo) Then Begin
      MemoFile := TFileStream.Create(FFileName + '.dbt', fmOpenReadWrite);
      MemoFile.Seek(0, soFromBeginning);
      MemoFile.ReadBuffer(DBTHeader, Sizeof(DBTHeader));
    End;
  End;
  If (Not FActive) And (IsOpen) Then Begin
    DBFile.Free;
    MemoFile.Free;
    IsOpen := False;
  End;
End;

Function TXBase.GetFieldName(FieldNo: Integer): String;
Var
  TmpC: Integer;
  TmpS: String; 
  
Begin
  If (FieldNo >= 1) And (FieldNo <= NumFields) Then Begin
    TmpC := 1;
    TmpS := '';
    While FieldStruct[FieldNo].FieldName[TmpC] <> #0 Do Begin
      TmpS := TmpS + FieldStruct[FieldNo].FieldName[TmpC];
      inc(TmpC);
    End;
    Result := UpperCase(TmpS);
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number: ' + IntToStr(FieldNo));
  End;
End;

Function TXBase.GetFieldType(FieldNo: Integer): TFieldType;
Begin
  If (FieldNo >= 1) And (FieldNo <= NumFields) Then Begin
    Result := xbfUnknown;
    Case ord(FieldStruct[FieldNo].FieldType) Of
      67: Result := xbfChar;
      68: Result := xbfDate;
      76: Result := xbfLogic;
      77: Result := xbfMemo;
      78: Result := xbfNumber;
    End;
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number: ' + IntToStr(FieldNo));
  End;
End;

Function TXBase.GetFieldSize(FieldNo: Integer): Integer;
Begin
  If (FieldNo >= 1) And (FieldNo <= NumFields) Then Begin
    Result := Ord(FieldStruct[FieldNo].FieldLength);
  End Else Begin
    Raise EXBaseException.Create('Invalid Field Number: ' + IntToStr(FieldNo));
  End;
End;

Function TXBase.Changing: Boolean;
Var
  Allowed: Boolean;
  
Begin
  Allowed := True;
  If Assigned(FOnChanging) Then Begin
    FOnChanging(Self, Allowed);
  End;
  Result := Allowed;
End;

Function TXBase.DoDelete: Boolean;
Var
  Allowed: Boolean;
  
Begin
  Allowed := True;
  If Assigned(FOnDeleting) Then Begin
    FOnDeleting(Self, Allowed);
  End;
  Result := Allowed;
End;

Procedure TXBase.AppendBlank;
Var
  TmpI: Integer;
  TmpC: Char;

Begin
  If IsOpen Then Begin
    inc(Header.NumRecords);
    Try
      DBFile.Seek((Header.HeaderLen) + ((Header.NumRecords - 1) * Header.RecordLen), soFromBeginning);
      TmpC := ' ';
      For TmpI := 1 To Header.RecordLen Do Begin
        DBFile.WriteBuffer(TmpC, 1);
      End;
      DbFile.Seek(0, soFromBeginning);
      DBFile.WriteBuffer(Header, Sizeof(Header));
      if Assigned(FOnAppended) then begin
        FOnAppended(Self);
      end;
    Except
      Raise EXBaseException.Create('Error Writing Record');
    End;

  End Else Begin
    Raise EXBaseException.Create('Data File Not Open');
  End;
End;

Function TXBase.IsBof: Boolean;
Begin
  Result := False;
  If FCurrentRecord = 1 Then Result := True;
End;

Function TXBase.IsEof: Boolean;
Begin
  Result := False;
  If FCurrentRecord = Header.NumRecords Then Result := True;
End;

Function TXBase.GetDeleteStatus: Boolean;
Begin
  If IsOpen Then Begin
    Result := (RecordBuffer[1] = '*');
  End Else Begin
    Result := False;
  End;
End;

Procedure TXBase.SetDeleteStatus(Delete: Boolean);
Begin
  If IsOpen Then Begin
    If Delete And DoDelete Then Begin
      RecordBuffer[1] := '*';
    End Else Begin
      RecordBuffer[1] := ' ';
    End;
  End;
End;

{******************************************************************************}
{* Property Editor Code                                                       *}
{******************************************************************************}

procedure TXBFilenameProperty.Edit;
var
  XBFileOpen: TOpenDialog;
begin
  XBFileOpen := TOpenDialog.Create(Nil);
  XBFileOpen.Filename := GetValue;
  XBFileOpen.Filter := 'xBase Files|*.DBF';
  XBFileOpen.Options := XBFileOpen.Options + [ofPathMustExist, ofFileMustExist];
  try
    if XBFileOpen.Execute then SetValue(XBFileOpen.Filename);
  finally
    XBFileOpen.Free;
  end;
end;

function TXBFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

{--------------------------------------}
{ Register                             }
{--------------------------------------}
Procedure Register;
Begin
  RegisterComponents('XBase',[TXBase]);
  RegisterPropertyEditor(TypeInfo(String), TXBase, 'FileName', TXBFileNameProperty);
End;

End.

